open Ast
open Bytecode

module StringMap = Map.Make(String)

(* Symbol table: Information about all the names in scope *)
type env = {
    function_index : int StringMap.t; (* Index for each function *)
    global_index   : int StringMap.t; (* "Address" for global variables *)
    local_index    : int StringMap.t; (* FP offset for args, locals *)
  }

(* val enum : int -> 'a list -> (int * 'a) list *)
let rec enum stride n = function
    [] -> []
  | hd::tl -> (n, hd) :: enum stride (n+stride) tl

(* val string_map_pairs StringMap 'a -> (int * 'a) list -> StringMap 'a *)
let string_map_pairs map pairs =
  List.fold_left (fun m (i, n) -> StringMap.add n i m) map pairs

(** Translate a program in AST form into a bytecode program.  Throw an
    exception if something is wrong, e.g., a reference to an unknown
    variable or function *)
let translate (globals, functions) =

  (* Allocate "addresses" for each global variable *)
  let global_indexes = string_map_pairs StringMap.empty (enum 1 0 (List.map (fun f -> f.vname)globals)) in

  (* Assign indexes to function names; built-in "print" is special *)
  let add_function map (a, b)=StringMap.add a b map in
  let built_in_functions = List.fold_left add_function StringMap.empty [("print",-1);("scan",-2);("random",-3);("printg",-4)] in
  let function_indexes = string_map_pairs built_in_functions
      (enum 1 1 (List.map (fun f -> f.fname) functions)) in

  (* Translate a function in AST form into a list of bytecode statements *)
  let translate env fdecl =
    (* Bookkeeping: FP offsets for locals and arguments *)
    let num_formals = List.length fdecl.formals
    and num_locals = List.length fdecl.locals
    and local_offsets = enum 1 1 (List.map (fun f -> f.vname) fdecl.locals)
    and formal_offsets = enum (-1) (-2) (List.map (fun f -> f.foname) fdecl.formals) in
    let env = { env with local_index = string_map_pairs
		  StringMap.empty (local_offsets @ formal_offsets) } in

    let rec expr = function
	Literal i -> [Lit i]
      | String j -> [Lits j]
      | Bool i -> if i=true then [Lit 1] else [Lit 0]
      | Array a -> [Lita a]
      | Array1 (a,e) -> expr e @ (try [Lfp (StringMap.find a env.local_index)]
				with Not_found -> try [Lod (StringMap.find a env.global_index)]
				with Not_found -> raise (Failure ("undeclared variable "^a)))
   (*   | Array2 (a,e1,e2)-> *)
      | Id s ->
	 [Lit 0] @ (try [Lfp (StringMap.find s env.local_index)]
          with Not_found ->  try [Lod (StringMap.find s env.global_index)]
          with Not_found -> raise (Failure ("undeclared variable " ^ s)))
      | Cell(e1,e2)-> expr e1 @ expr e2 @ [Lodg]
      | Binop (e1, op, e2) -> expr e1 @ expr e2 @ [Bin op]
      | Assign (s, e) -> (expr e) @ (match s with
		Id s1 -> [Lit 0] @ (try [Sfp (StringMap.find s1 env.local_index)]
  	  with Not_found -> try [Str (StringMap.find s1 env.global_index)]
	  with Not_found -> raise (Failure ("undeclared variable " ^ s1)))
		| Array1 (a,e)->(expr e) @(try [Sfp (StringMap.find a env.local_index)]
  	  with Not_found -> try [Str (StringMap.find a env.global_index)]
	  with Not_found -> raise (Failure ("undeclared variable " ^ a)))
		| Cell(e1,e2)->((expr e1) @ (expr e2) @ [Strg])
		| _ -> raise (Failure("unreasonable expression")))
      | Not e -> expr e @ [Nota]
      | SumRow e -> expr e @ [Lit 0; Lit (-12); Strg]
      | SumCol e -> expr e @ [Lit 0; Lit (-11); Strg]
      | SumDiag e -> expr e @ [Lit 0; Lit (-10); Strg]
      | SumBlock e -> expr e @ [Lit 0; Lit (-9); Strg]
      | MaxRow e -> expr e @ [Lit 0; Lit (-8); Strg]
      | MaxCol e -> expr e @ [Lit 0; Lit (-7); Strg]
      | MaxDiag e -> expr e @ [Lit 0; Lit (-6); Strg]
      | MaxBlock e -> expr e @ [Lit 0; Lit (-5); Strg]
      | MinRow e -> expr e @ [Lit 0; Lit (-4); Strg]
      | MinCol e -> expr e @ [Lit 0; Lit (-3); Strg]
      | MinDiag e -> expr e @ [Lit 0; Lit (-2); Strg] 
      | MinBlock e -> expr e @ [Lit 0; Lit (-1); Strg]
      | IsVisible (e1, e2) -> expr e1 @ expr e2 @ [Setv]
      | CheckGrid -> [Chkg]
      | Call (fname, actuals) -> (try
	  (List.concat (List.map expr (List.rev actuals))) @
	  [Jsr (StringMap.find fname env.function_index) ]   
        with Not_found -> raise (Failure ("undefined function " ^ fname)))
      | Noexpr -> []

    in let rec stmt = function
	Block sl     ->  List.concat (List.map stmt sl)
      | Expr e       -> expr e @ [Drp]
      | Return e     -> expr e @ [Rts num_formals]
      | IfElse (p, t, f) -> let t' = stmt t and f' = stmt f in
	expr p @ [Beq(2 + List.length t')] @
	t' @ [Bra(1 + List.length f')] @ f'
      | For (e1, e2, e3, b) ->
	  stmt (Block([Expr(e1); While(e2, Block([b; Expr(e3)]))]))
      | While (e, b) ->
	  let b' = stmt b and e' = expr e in
	  [Bra (1+ List.length b')] @ b' @ e' @
	  [Bne (-(List.length b' + List.length e'))]
      | Nostmt -> []

    in [Ent num_locals] @      (* Entry: allocate space for locals *)
    stmt (Block fdecl.body) @  (* Body *)
    [Lit 0; Rts num_formals]   (* Default = return 0 *)

  in let env = { function_index = function_indexes;
		 global_index = global_indexes;
		 local_index = StringMap.empty } in

  (* Code executed to start the program: Jsr main; halt *)
  let entry_function = try
    [Jsr (StringMap.find "main" function_indexes); Hlt]
  with Not_found -> raise (Failure ("no \"main\" function"))
  in
    
  (* Compile the functions *)
  let func_bodies = entry_function :: List.map (translate env) functions in

  (* Calculate function entry points by adding their lengths *)
  let (fun_offset_list, _) = List.fold_left
      (fun (l,i) f -> (i :: l, (i + List.length f))) ([],0) func_bodies in
  let func_offset = Array.of_list (List.rev fun_offset_list) in

  { num_globals = List.length globals;
    (* Concatenate the compiled functions and replace the function
       indexes in Jsr statements with PC values *)
    text = Array.of_list (List.map (function
	Jsr i when i > 0 -> Jsr func_offset.(i)
      | _ as s -> s) (List.concat func_bodies))
  }
