(* Simple binary operators *)
type op =
Add | Sub | Mult | Div | Equal | Neq | Less |
Leq | Greater | Geq | And | Or 

type t=
Integer|Strings|Boolean|Arrays

type expr = 
	 Id of string
	|Literal of int
	|Bool of bool
	|String of string
	|Array of int list
	|Array1 of string * expr
  | Binop of expr * op * expr
  | Not of expr
  | Assign of expr * expr
  | Cell of expr * expr
  | Row of expr
  | Column of expr
  | Random of expr 
  | IsVisible of expr * expr
  | SumRow of expr
  | SumCol of expr
  | SumDiag of expr
  | MaxRow of expr
  | MaxCol of expr
  | MaxDiag of expr
  | MinRow of expr
  | MinCol of expr
  | MinDiag of expr 
  | CheckGrid
  | Call of string * expr list
  | Noexpr


type decl={
 vtype: t;
 vname: string;
}

type formal={
ftype: t;
foname: string;
}

type stmt = 
  | IfElse of expr * stmt * stmt 
  | While of expr * stmt
  | For of expr * expr* expr * stmt
  | Expr of expr
  | Return of expr
  | Block of stmt list
  | Nostmt

type func_decl = {
  fname : string ;
  formals : formal list ;
  locals :  decl list ;
  body   :  stmt list   ;
  }

type program = decl list * func_decl list (* starting point *)

let rec string_of_array=function
		[]->""
		| [b]->string_of_int b
		| hd::tl->string_of_int hd^", "^string_of_array tl
	

let rec string_of_expr = function
	Literal(l) -> string_of_int l
	| Bool(true) -> "true"
	| Bool(false) -> "false"
	| String(s) ->"\"" ^ s ^"\""
	| Array(a) -> "[| "^
		string_of_array a
		^" |]"
	| Array1(a,e)-> a^"[ "^string_of_expr e^" ]"
	| Id(b)-> b
	| Binop(e1, o, e2) ->
	   string_of_expr e1^" "^
	    (match o with
		Add->"+" | Sub -> "-"| Mult -> "*" | Div -> "/"
		| Equal -> "=" | Neq -> "!=" | Greater -> ">" 
		| Less -> "<" | Leq -> "<=" | Geq -> ">=" | And -> "and" | Or -> "or")^" "
		^string_of_expr e2
	| Assign(e1,e2)->string_of_expr e1^"~"^string_of_expr e2
	| Not(e1)->"not "^string_of_expr e1
	| Cell(e1, e2)->"Grid.Cell("^string_of_expr e1^", "^string_of_expr e2^")"
	| Row(e)->"Grid.Row("^string_of_expr e^")"
	| Column(e)->"Grid.Column("^string_of_expr e^")"
	| Random(e)->"Random ("^string_of_expr e^")" 
	| IsVisible(e1,e2)->"Grid.MakeVisible("^string_of_expr e1^", "^string_of_expr e2^")"
	| SumRow(e)->"Grid.SumofRows("^string_of_expr e^")"
	| SumCol(e)->"Grid.SumOfColumns("^string_of_expr e^")"
	| SumDiag(e)->"Grid.SumOfDiagnals("^string_of_expr e^")"
	| MaxRow(e)->"Grid.MaxValueOfRows("^string_of_expr e^")"
	| MaxCol(e)->"Grid.MaxValueOfColumns("^string_of_expr e^")"
	| MaxDiag(e)->"Grid.MaxValueOfDiagnals("^string_of_expr e^")"
	| MinRow(e)->"Grid.MinValueOfRows("^string_of_expr e^")"
	| MinCol(e)->"Grid.MinValueOfColumns("^string_of_expr e^")"
	| MinDiag(e)->"Grid.MinValueOfDiagnals("^string_of_expr e^")"
	| CheckGrid -> "Grid.CheckGrid()"
	| Call(f,e1) -> f^"( "^String.concat ", " (List.map string_of_expr e1)^" )"
	| Noexpr -> ""

let rec string_of_stmt = function
	 Block(stmts) -> "begin\n"^String.concat "" (List.map string_of_stmt stmts)^"end\n"
	| Expr(expr) -> string_of_expr expr^";\n";
	| Return(expr) -> "return " ^string_of_expr expr^";\n";
	| IfElse(expr, stmt, Block([])) -> "if( "^string_of_expr expr^" )\n"^string_of_stmt stmt
	| IfElse(expr,stmt1,stmt2) -> "if( "^string_of_expr expr^" )\n"^string_of_stmt stmt1^
		"else\n"^string_of_stmt stmt2
	| For(e1,e2,e3,stmt) -> "for( "^string_of_expr e1^"; "^string_of_expr e2^"; "^string_of_expr e3^" )\n"^string_of_stmt stmt
	| While(expr,stmt) -> "while( "^string_of_expr expr^" )\n"^string_of_stmt stmt
	| Nostmt->""

let string_of_t = function
	Integer->"int"
	| Strings -> "string" 
	| Boolean -> "bool"
	| Arrays -> "Array"

let string_of_vdecl vdecl=string_of_t vdecl.vtype ^ " " ^ vdecl.vname^";\n"

let string_of_formal formal=string_of_t formal.ftype ^ " " ^ formal.foname

let string_of_fdecl fdecl=
	fdecl.fname^" ("^String.concat ", " (List.map string_of_formal fdecl.formals)^ " )\nbegin\n"^String.concat "" (List.map string_of_vdecl fdecl.locals)^String.concat "" (List.map string_of_stmt fdecl.body)^"end\n"

let string_of_program (vars, funcs)=
	String.concat "" (List.map string_of_vdecl vars)^"\n"^
	String.concat "\n" (List.map string_of_fdecl funcs)
